# Generated by Django 2.0.13 on 2019-06-13 11:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0024_layer_geom_type'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['layer'], name='geo_feat_layer_i_921f95_idx'),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['updated_at'], name='geo_featu_updated_62a19a_idx'),
        ),
    ]



from django.urls import reverse
from django.utils.http import urlunquote
from rest_framework import serializers
from rest_framework.fields import empty

from geostore.models import (Feature, FeatureRelation, Layer,
                             LayerRelation, LayerGroup)
from geostore.validators import (validate_json_schema_data,
                                 validate_json_schema, validate_geom_type)


class FeatureSerializer(serializers.ModelSerializer):
    properties = serializers.JSONField(required=False)

    def __init__(self, instance=None, data=empty, **kwargs):
        super().__init__(instance=instance, data=data, **kwargs)
        self.layer = None

    def get_layer(self):
        if self.instance:
            self.layer = self.instance.layer
        if not self.layer and self.context.get('layer_pk'):
            self.layer = Layer.objects.get(pk=self.context.get('layer_pk'))
        return self.layer

    def validate_geom(self, data):
        """
        Validate geom exists
        """
        if self.get_layer():
            validate_geom_type(self.get_layer().geom_type, data.geom_typeid)
        return data

    def validate_properties(self, data):
        """
        Validate schema if exists
        """
        if self.get_layer():
            validate_json_schema_data(data, self.get_layer().schema)
        return data

    class Meta:
        model = Feature
        fields = ('id', 'identifier', 'layer', 'geom', 'properties', )
        read_only_fields = ('id', 'layer')


class GroupSerializer(serializers.ModelSerializer):
    group_tilejson = serializers.SerializerMethodField()
    group_tiles = serializers.SerializerMethodField()

    def get_group_tilejson(self, obj):
        return urlunquote(reverse('geostore:group-tilejson', args=[obj.slug]))

    def get_group_tiles(self, obj):
        return urlunquote(reverse('geostore:group-tiles-pattern', args=[obj.slug]))

    class Meta:
        model = LayerGroup
        fields = '__all__'


class LayerSerializer(serializers.ModelSerializer):
    routing_url = serializers.SerializerMethodField()
    shapefile_url = serializers.SerializerMethodField()
    geojson_url = serializers.SerializerMethodField()
    schema = serializers.JSONField(required=False, validators=[validate_json_schema])
    layer_intersects = serializers.SerializerMethodField()
    layer_tilejson = serializers.SerializerMethodField()
    layer_groups = GroupSerializer(many=True, read_only=True)

    def get_routing_url(self, obj):
        return reverse('geostore:layer-route', args=[obj.pk, ])

    def get_shapefile_url(self, obj):
        return reverse('geostore:layer-shapefile', args=[obj.pk, ])

    def get_geojson_url(self, obj):
        return reverse('geostore:layer-geojson', args=[obj.pk, ])

    def get_layer_intersects(self, obj):
        return reverse('geostore:layer-intersects', args=[obj.name, ])

    def get_layer_tilejson(self, obj):
        return urlunquote(reverse('geostore:layer-tilejson', args=[obj.pk]))

    class Meta:
        model = Layer
        fields = '__all__'


class GeoJSONLayerSerializer(serializers.JSONField):
    def to_representation(self, data):
        return data.to_geojson()


class LayerRelationSerializer(serializers.ModelSerializer):
    class Meta:
        model = LayerRelation
        fields = ('id', 'origin', 'destination', 'schema')


class FeatureRelationSerializer(serializers.ModelSerializer):
    class Meta:
        model = FeatureRelation
        fields = ('id', 'origin', 'destination')



import io
import logging
import os
import zipfile

from django.contrib.gis.geos.point import Point

logger = logging.getLogger(__name__)


def make_zipfile_bytesio(base_dir):
    zip_file = io.BytesIO()

    with zipfile.ZipFile(zip_file, "w",
                         compression=zipfile.ZIP_DEFLATED) as zf:
        path = os.path.normpath(base_dir)
        if path != os.curdir and path != base_dir:
            zf.write(path, os.path.relpath(path, base_dir))
            logger.info("adding '%s'", path)
        for dirpath, dirnames, filenames in os.walk(base_dir):
            for name in sorted(dirnames):
                path = os.path.normpath(os.path.join(dirpath, name))
                zf.write(path, os.path.relpath(path, base_dir))
                logger.info("adding '%s'", path)
            for name in filenames:
                path = os.path.normpath(os.path.join(dirpath, name))
                if os.path.isfile(path):
                    zf.write(path, os.path.relpath(path, base_dir))
                    logger.info("adding '%s'", path)

    return zip_file


class ChunkIterator:

    def __init__(self, iterator, chunksize):
        self.iterator = iterator
        self.chunksize = chunksize

    def __iter__(self):
        return self

    def __next__(self):
        try:
            chunk = []
            for i in range(self.chunksize):
                chunk.append(next(self.iterator))
        finally:
            if chunk:
                return chunk
            else:
                raise StopIteration

    def next(self):
        return self.__next__()


class GeometryDefiner:
    LONGITUDE = 'longitude'
    LATITUDE = 'latitude'

    @staticmethod
    def get_geometry(column_names, row):
        if not isinstance(column_names, dict):
            return None
        if sorted(column_names.keys()) == [GeometryDefiner.LATITUDE,
                                           GeometryDefiner.LONGITUDE]:
            lat_column = column_names.get(GeometryDefiner.LATITUDE)
            long_column = column_names.get(GeometryDefiner.LONGITUDE)
            if all(row.get(column) for column in [long_column, lat_column]):
                x = float(row.get(long_column))
                y = float(row.get(lat_column))
                return Point(x, y)
        return None



from django.contrib.postgres.fields.jsonb import JSONField
from django.core.exceptions import FieldDoesNotExist
from django.db.models import Q
from rest_framework.filters import BaseFilterBackend


class JSONFieldFilterBackend(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        query = Q()
        for param_name, param_value in request.query_params.items():
            try:
                field = (queryset.model
                                 ._meta
                                 .get_field(param_name.split('__')[0]))
            except FieldDoesNotExist:
                pass
            else:
                if isinstance(field, JSONField):
                    sub_query = Q(**{param_name: param_value})
                    try:
                        sub_query |= Q(**{param_name: int(param_value)})
                    except ValueError:
                        pass
                    query &= sub_query
        return queryset.filter(query)



from django.core.management.base import BaseCommand, CommandError
from django.db import connection, transaction
from django.utils.module_loading import import_string

from geostore.management.commands.mixins import LayerCommandMixin
from geostore.models import Feature, Layer


class Command(LayerCommandMixin, BaseCommand):
    help = 'Run a data processing on layers - UNSAFE command'

    def add_arguments(self, parser):
        # Inputs Layers
        # pk and name are mutually exclusive to avoid supporting the order of this two kind of parameters.
        exclusive_group_layer_ins = parser.add_mutually_exclusive_group(required=True)
        exclusive_group_layer_ins.add_argument(
            '-pk-ins', '--layer-pk-ins',
            default=[],
            action='append',
            help=("PKs of the input layers."))
        exclusive_group_layer_ins.add_argument(
            '-name-ins', '--layer-name-ins',
            default=[],
            action='append',
            help=("Names of the input layers."))

        # Output Layers
        exclusive_group_layer_out = parser.add_mutually_exclusive_group()
        exclusive_group_layer_out.add_argument(
            '-pk-out', '--layer-pk-out',
            type=int,
            action="store",
            help=("PK of the output layer."))
        exclusive_group_layer_out.add_argument(
            '-name-out', '--layer-name-out',
            action="store",
            help=("Name of the output layer."))

        parser.add_argument(
            '-co', '--clear_output',
            action="store_true",
            help=("Empty the output layer before adding data."))

        # Processing action
        exclusive_group_processing = parser.add_mutually_exclusive_group(required=True)
        exclusive_group_processing.add_argument(
            '-s', '--sql',
            action="store",
            help=("UNSAFE SQL Query. Eg. SELECT identifier, properties, ST_Centroid(geom::geometry) AS geom FROM in0"))
        exclusive_group_processing.add_argument(
            '-s-centroid', '--sql-centroid',
            action="store_true",
            help=("Compute centroid"))
        exclusive_group_processing.add_argument(
            '-p', '--python',
            help=("UNSAFE Compute with a python callable module, use multiple key=value as arguments"))
        exclusive_group_processing.add_argument(
            '-mv', '--make-valid',
            action="store_true",
            help=("Enforce valide geomtries"))

        parser.add_argument(
            'command_args',
            nargs='*',
            action="store",
            help='Processing command arguments')

        parser.add_argument(
            '--dry-run',
            action="store_true",
            help='Execute une dry-run mode')

    def _get_layer_ins(self, pks, names):
        try:
            return ([Layer.objects.get(id=id) for id in pks] +
                    [Layer.objects.get(name=name) for name in names])
        except Layer.DoesNotExist:
            raise CommandError(f"Fails open one or many layers layer-pk-ins: {', '.join(pks)} "
                               f"and layer-name-ins: {', '.join(names)}")

    def _get_layer_out(self, pk, name, clear_output, verbosity):
        if pk or name:
            layer = self._get_layer_by_pk(pk) if pk else self._get_layer_by_name(name)
            if clear_output:
                layer.features.all().delete()
        else:
            layer = Layer.objects.create()
            if verbosity >= 1:
                self.stdout.write(
                    f"The created layer pk is {layer.pk} and this name \"{layer.name}\""
                    "it can be used to import more features in the same layer with different options")

        return layer

    @transaction.atomic()
    def handle(self, *args, **options):
        dryrun = options.get('dry_run')
        sp = transaction.savepoint()
        layer_ins = self._get_layer_ins(
            options.get('layer_pk_ins'),
            options.get('layer_name_ins')
        )
        layer_out = self._get_layer_out(
            options.get('layer_pk_out'),
            options.get('layer_name_out'),
            options.get('clear_output'),
            options['verbosity']
        )
        command_args = dict([a.split('=', 1) for a in options.get('command_args') or []])

        sql = options.get('sql')
        python_object_name = options.get('python')

        if python_object_name:
            self._call(python_object_name, layer_ins, layer_out, **command_args)
        elif options.get('sql_centroid'):
            self._simple_sql('ST_Centroid', layer_ins, layer_out)
        elif sql:
            self._sql(sql, layer_ins, layer_out)
        elif options.get('make_valid'):
            self._processing_make_valid(layer_ins, layer_out)
        if dryrun:
            transaction.savepoint_rollback(sp)
        else:
            transaction.savepoint_commit(sp)

    def _call(self, python_callable_name, layer_ins, layer_out, **command_args):
        callable_object = import_string(python_callable_name)
        callable_object(layer_ins, layer_out, **command_args)

    def _simple_sql(self, sql_function, layer_ins, layer_out):
        return self._sql(
            f'SELECT identifier, properties, {sql_function}(geom::geometry) AS geom FROM in0',
            layer_ins, layer_out)

    def _sql(self, sql, layer_ins, layer_out):
        args = []
        raws = []
        for (i, l) in enumerate(layer_ins):
            raw, arg = l.features.all().query.sql_with_params()
            raws.append(f'in{i} AS ({raw})')
            args.append(arg)
        with_ = ',\n'.join(raws)

        with connection.cursor() as cursor:
            sql_query = f'''
                WITH
                {with_}
                INSERT INTO {Feature._meta.db_table} (layer_id, created_at, updated_at, identifier, properties, geom)
                SELECT {layer_out.id}, NOW(), NOW(), * FROM (
                    {sql}
                ) AS t
            '''

            cursor.execute(sql_query, args)

    def _processing_make_valid(self, layer_ins, layer_out):
        if len(layer_ins) != 1:
            raise ValueError('Exactly one input layer required')
        layer_in = layer_ins[0]

        if layer_in.is_point:
            raise NotImplementedError
        elif layer_in.is_linestring:
            raise NotImplementedError
        elif layer_in.is_polygon:
            if not layer_in.is_multi:
                # Polygon
                self._sql(
                    """
                    SELECT
                        identifier,
                        properties,
                        CASE
                            WHEN (
                                SELECT geom
                                FROM ST_Dump(ST_CollectionExtract(ST_MakeValid(geom::geometry), 3))
                                LIMIT 1
                            ) IS NOT NULL
                                THEN (
                                    SELECT geom
                                    FROM ST_Dump(ST_CollectionExtract(ST_MakeValid(geom::geometry), 3))
                                    LIMIT 1
                                )
                            ELSE ST_BUFFER(geom::geometry, 0)
                        END
                    FROM
                        in0
                    """,
                    layer_ins, layer_out)
            else:
                # MultiPolygon
                self._sql(
                    """
                    SELECT
                        identifier,
                        properties,
                        CASE
                            WHEN ST_CollectionExtract(ST_MakeValid(geom::geometry), 3) IS NOT NULL
                                THEN ST_CollectionExtract(ST_MakeValid(geom::geometry), 3)
                            ELSE ST_BUFFER(geom::geometry, 0)
                        END
                    FROM
                        in0
                    """,
                    layer_ins, layer_out)



# Generated by Django 2.1.10 on 2019-07-24 09:01

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0027_layergroup'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='layer',
            name='group',
        ),
    ]



from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.http import Http404


class MultipleFieldLookupMixin(object):
    """
    Apply this mixin to any view or viewset to get multiple field filtering
    based on a `lookup_fields` attribute, instead of the default single field
    filtering.
    """
    def get_object(self):

        queryset = self.get_queryset()             # Get the base queryset
        queryset = self.filter_queryset(queryset)  # Apply any filter backends

        value = self.kwargs[self.lookup_field]
        for field in self.lookup_fields:
            try:
                obj = queryset.get(**{field: value})
            except (ObjectDoesNotExist, ValueError):
                continue
            else:
                self.check_object_permissions(self.request, obj)
                return obj
        raise Http404


class BaseUpdatableModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True



import datetime
import re

from django.core import exceptions
from django.db.models import DateField
from django.utils.dateparse import parse_date
from django.utils.translation import gettext_lazy as _


class DateFieldYearLess(DateField):
    base_year = 1970
    default_error_messages = {
        'invalid': _("'%(value)s' value has an invalid date format. It must be"
                     " in MM-DD format."),
        'invalid_date': _("'%(value)s' value has the correct format (MM-DD) "
                          "but it is an invalid date."),
    }

    def from_db_value(self, value, expression, connection, context):
        return f"{value:%m-%d}"

    def to_python(self, value):
        if isinstance(value, datetime.date):
            value = f'{value:%m-%d}'
        if not re.search(r'^\d{2}-\d{2}$', value):
            raise exceptions.ValidationError(
                self.error_messages['invalid'],
                code='invalid',
                params={'value': value},
            )

        normalized_date = parse_date(f'1970-{value}')
        return super().to_python(normalized_date)



import argparse
import json
import uuid
from json.decoder import JSONDecodeError

from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from geostore.management.commands.mixins import LayerCommandMixin
from geostore.models import Layer, LayerGroup


class Command(LayerCommandMixin, BaseCommand):
    help = 'Import Features from Zipped ShapeFile'

    def add_arguments(self, parser):
        parser.add_argument('file_path',
                            nargs='+',
                            type=argparse.FileType('rb'),
                            action="store",
                            help='Ziped ShapeFile files to import')

        exclusive_group = parser.add_mutually_exclusive_group()
        exclusive_group.add_argument('-pk', '--layer-pk',
                                     type=int,
                                     action="store",
                                     help=("PK of the layer where to insert"
                                           "the features.\n"
                                           "A new layer is created if not "
                                           "present."))
        exclusive_group.add_argument('-ln', '--layer-name',
                                     action="store",
                                     help=("Name of created layer "
                                           "containing ShapeFile datas."
                                           "If not provided an uuid4 is set."))
        parser.add_argument('-gs', '--generate-schema',
                            action='store_true',
                            help=("Generate JSON schema from "
                                  "ShapeFile properties.\n"
                                  "Only available if not -pk option provided."))
        parser.add_argument('-ls', '--layer-settings', nargs='?',
                            type=argparse.FileType('r'),
                            action="store",
                            help="JSON settings file to override default")

        parser.add_argument('-i', '--identifier',
                            action="store",
                            help="Field in properties that will be used as "
                                 "identifier of the features, so features can"
                                 " be grouped on layer's operations")
        parser.add_argument('-gr', '--group',
                            action="store",
                            default="default",
                            help="Group name of the created layer")
        parser.add_argument('--dry-run',
                            action="store_true",
                            help='Execute une dry-run mode')

    @transaction.atomic()
    def handle(self, *args, **options):
        layer_pk = options.get('layer_pk')
        layer_name = options.get('layer-name') or uuid.uuid4()
        file_path = options.get('file_path')
        dryrun = options.get('dry_run')
        group = options.get('group')
        identifier = options.get('identifier')
        generate_schema = options.get('generate_schema')
        sp = transaction.savepoint()

        if layer_pk:
            layer = self._get_layer_by_pk(layer_pk)
        else:
            try:
                layer_settings = options.get('layer_settings')
                settings = json.loads(layer_settings.read()) if layer_settings else {}
            except (JSONDecodeError, UnicodeDecodeError):
                raise CommandError("Please provide a valid layer settings file")

            layer = Layer.objects.create(name=layer_name, settings=settings)
            if group:
                group, created = LayerGroup.objects.get_or_create(name=group)
                group.layers.add(layer)

            if options['verbosity'] > 0:
                self.stdout.write(
                    f"The created layer pk is {layer.pk}, "
                    "it can be used to import more features "
                    "in the same layer with different "
                    "options"
                )

        self.import_datas(layer, file_path, identifier)
        if generate_schema and not layer_pk:
            # only in layer creation, find properties to generate schema
            layer.schema = {
                'properties': {
                    key: {
                        'type': 'string'
                    } for key, value in layer.layer_properties.items()
                }
            }
            layer.save()

        if dryrun:
            transaction.savepoint_rollback(sp)
        else:
            transaction.savepoint_commit(sp)

    def import_datas(self, layer, file_path, identifier):
        for shapefile_file in file_path:
            layer.from_shapefile(shapefile_file, identifier)



import logging

from .helpers import GeometryDefiner

logger = logging.getLogger(__name__)


def set_geometry_from_options(feature_args, options):
    geometry_columns = {
        GeometryDefiner.LONGITUDE: options.get('longitude'),
        GeometryDefiner.LATITUDE: options.get('latitude')
    }
    geometry_columns_filtered = {k: v for k, v in geometry_columns.items()
                                 if v is not None}

    geometry = GeometryDefiner.get_geometry(
        column_names=geometry_columns_filtered,
        row=feature_args["properties"]
    )
    if geometry:
        feature_args['geom'] = geometry
    else:
        pk_properties = [(key, feature_args["properties"].get(key)) for key in
                         options.get('pk_properties')]
        logger.warning(f'can not define geometry for: {str(pk_properties)}')



# Generated by Django 2.0.9 on 2018-11-14 08:06

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0016_auto_20181120_1059'),
    ]

    operations = [
        migrations.AddField(
            model_name='layer',
            name='settings',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict),
        ),
    ]



from django.contrib.auth.models import Permission
from django.contrib.gis import admin

from . import models


@admin.register(models.Layer)
class LayerAdmin(admin.ModelAdmin):
    list_display = ('pk', 'name', 'geom_type', 'layer_groups')
    list_filter = ('geom_type', 'layer_groups')


@admin.register(models.LayerGroup)
class LayerGroupAdmin(admin.ModelAdmin):
    list_display = ('name', 'slug')
    list_filter = ('slug', )


@admin.register(models.Feature)
class FeatureAdmin(admin.OSMGeoAdmin):
    list_display = ('pk', 'identifier', 'layer', 'source', 'target')
    list_filter = ('layer', )


admin.site.register(models.FeatureRelation)
admin.site.register(Permission)



# Generated by Django 2.0.13 on 2019-04-15 09:45

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0020_auto_20190401_1139'),
    ]

    operations = [
        migrations.AddField(
            model_name='feature',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='feature',
            name='updated_at',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='layer',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='layer',
            name='updated_at',
            field=models.DateTimeField(auto_now=True),
        ),
    ]
